package com.sample;

import java.io.InputStreamReader;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.drools.KnowledgeBase;
import org.drools.KnowledgeBaseFactory;
import org.drools.StatefulSession;
import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderError;
import org.drools.builder.KnowledgeBuilderErrors;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.ResourceType;
import org.drools.io.ResourceFactory;
import org.drools.runtime.StatefulKnowledgeSession;

/**
 * This is a sample class to launch a rule.
 */
public class DroolsTest {
	
	public static final void main(String[] args) throws Exception {
		new DroolsTest().go(true);
	}
	
	public void go(boolean fireUntilHaltTest) throws Exception {

		// load up the knowledge base
		KnowledgeBase kbase = readKnowledgeBase();
		
		StatefulKnowledgeSession ksession = kbase.newStatefulKnowledgeSession();

		long time = System.currentTimeMillis();
		for(Named n : getSetupData()) {
			ksession.insert(n);
		}
		if(fireUntilHaltTest) {
			ksession.fireUntilHalt();
		} else {
			ksession.fireAllRules();
		}
		long elapsed = System.currentTimeMillis() - time;

		System.out.println("Time: " + elapsed);
		System.out.println();
	}
	
	protected Collection<Named> getSetupData() {
		
		List<Named> data = new LinkedList<Named>();
		
		Location p1 = new Location("p1"); data.add(p1);
		Location p2 = new Location("p2"); data.add(p2);
		Location p3 = new Location("p3"); data.add(p3);
		
		Block b1 = new Block("b1"); p1.setHolding(b1); data.add(b1);
		Block b2 = new Block("b2"); p2.setHolding(b2); data.add(b2);
		Block b3 = new Block("b3"); p3.setHolding(b3); data.add(b3);
		Block b4 = new Block("b4"); b3.setHolding(b4); data.add(b4);
		
		p1.setHolding(b1);
		
		data.add(new MovementGoal("Move b1 to p1", b1, p1));
		
		return data;
	}
	
	private static KnowledgeBase readKnowledgeBase() throws Exception {

		long time = System.currentTimeMillis();
		
		KnowledgeBuilder kbuilder = KnowledgeBuilderFactory.newKnowledgeBuilder();
		kbuilder.add(ResourceFactory.newReaderResource(new InputStreamReader(DroolsTest.class.getResourceAsStream("/Sample.drl"))), ResourceType.DRL);
		KnowledgeBuilderErrors errors = kbuilder.getErrors();
		if (errors.size() > 0) {
			for (KnowledgeBuilderError error: errors) {
				System.err.println(error);
			}
			throw new IllegalArgumentException("Could not parse knowledge.");
		}
		
		KnowledgeBase kbase = KnowledgeBaseFactory.newKnowledgeBase();
		
		kbase.addKnowledgePackages(kbuilder.getKnowledgePackages());
		
		System.out.println("Rules load time: " + (System.currentTimeMillis() - time) + "ms" );
		
		return kbase;
	}
	
	public static class Named {
		
		public final String name;

		public String getName() {
			return name;
		}

		public Named(String name) {
			super();
			this.name = name;
		}
		
	}
	
	public static interface Thing {
		String getName();
	}
	
	public static interface Holder {
		String getName();
		Thing getHolding();
		void setHolding(Thing thing);
	}
	
	public static class Block extends Named implements Thing, Holder {

		public Block(String name) {
			super(name);
		}

		public Thing thing = null;

		@Override
		public Thing getHolding() {
			return thing;
		}
		
		public void setHolding(Thing thing) {
			this.thing = thing;
		}

		@Override
		public String toString() {
			return String.format("Block %s" + (thing == null ? "" : " (holding %s)"), name, thing);
		}
		
	}
	
	public static class Location extends Named implements Holder {

		public Location(String name) {
			super(name);
		}

		public Thing thing = null;
		
		@Override
		public Thing getHolding() {
			return thing;
		}
		
		public void setHolding(Thing thing) {
			this.thing = thing;
		}

		@Override
		public String toString() {
			return String.format("Location %s" + (thing == null ? "" : " (holding %s)"), name, thing);
		}
		
	}
	
	public static class Goal extends Named {

		public Goal(String name) {
			super(name);
		}

		@Override
		public String toString() {
			return String.format("Goal [name=%s]", name);
		}
	}
	
	public static class MovementGoal extends Goal {

		public final Thing thing;
		public final Holder to;

		public MovementGoal(String name, Thing thing, Holder to) {
			super(name);
			this.thing = thing;
			this.to = to;
		}

		public Thing getThing() {
			return thing;
		}

		public Holder getTo() {
			return to;
		}

		@Override
		public String toString() {
			return String.format(
					"MovementGoal [thing=%s, to=%s, name=%s]", thing,
					to, name);
		}
		
	}

}